//---------------------------------------------------------------------------
// GPGee - GNU Privacy Guard Explorer Extension
// GPGeeVerifyDecrypt.cpp - Verification form code
// Copyright 2005, Kurt Fitzner <kfitzner@excelcia.org>
//---------------------------------------------------------------------------
// This file is part of GPGee.
//
// GPGee is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// GPGee is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// File Notes:
// This is the file verification and decryption form.  Both the UI code and
// operational code are here.
//
// Unlike the sign/encrypt code which performs a single operation (although
// it can do it on multiple files), this code attempts to analyze each file
// to determine the correct operation to perform.  Thus it can verify a
// detached signature, an attached one, pk decrypt, symmetrical decrypt, etc.
// The key to this is determining what operation to perform.  This is
// dependent on what type of gpg file each file is.  The determination of the
// type is actually done externally by a function in the GPGeeUtility.cpp
// unit.  It gives us either GPG_SIGNATURE or GPG_MESSAGE.  The former is a
// detached signature - it is fed to gpgme_op_file_verify().  The latter can
// be many different things - an encrypted file (symmetrically or pk), or a
// standalone signed file (binary or clearsigned).  While it would be nice
// to further categorize the type (so we don't end up just telling the user
// it's a "message"), it isn't actually necesary since GnuPG will handle any
// of those with the --decrypt switch.  That means that the
// gpgme_op_file_decrypt() function can handle them.
//
// For this form we set up a passphrase callback function.  This is what lets
// us handle a standalone signed-file and an encrypted file the same way.  We
// don't care which of the two it is - if it's an ecnrypted file that needs
// a passphrase, the callback function will get called and we can ask for the
// passphrase at that time.  There is no password caching done, which is
// something we might want to think about in the future.  My preference is
// that we get some code in GPGME that will allow us to figure out exactly
// what sort of file any particular gpg file is.  This way we wouldn't need
// to do much password caching at all.  We could sort the files by their type
// and do all the operations that need a password for a given key at the same
// time - so only that one passphrase would need caching.  It would also
// allow us to proactively ask for a passphrase rather than needing to do it
// on request from the callback.  This would eliminate some issues with the
// callback (see the noted for the callback function).
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TConfiguration.h"
#include "GPGeeExceptions.h"
#include "GPGeePassPhrase.h"
#include "GPGeeVerifyDecrypt.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TformGPGeeVerifyDecrypt *formGPGeeVerifyDecrypt;
extern TConfiguration *GPGeeConfig;
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Form constructor
//
__fastcall TformGPGeeVerifyDecrypt::TformGPGeeVerifyDecrypt(TComponent* Owner): TForm(Owner)
{
  gpgme_error_t err;
  AnsiString sConfigString;

  ctxGPG = NULL;
  cacheSecret = NULL;
  cachePublic = NULL;

  Cancel = false;

  // Initialize the GPGME library main and keycache contexts
  err = gpgme_new(&ctxGPG);
  if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_LIB_INIT, gpgme_strerror(err)));

  // Cache the secret keys - this cache is used when a passphrase for one of these keys is needed - then we get all the
  // info about the key from the cache to present it to the user.  The public keys aren't cached here because they are
  // so seldomly needed.
  err = gpgme_keycache_new (&cacheSecret, ctxGPG);
  if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_INIT_SEC,gpgme_strerror(err)));
  err = gpgme_keycache_init(cacheSecret, NULL, true);
  if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_LOAD_SEC,gpgme_strerror(err)));

  // Set the passphrase callback function.  This can't be a method because of C++ limitations (bah), so we supply a
  // pointer to this form to the callback so that a class method can be invoked from the callback.
  gpgme_set_passphrase_cb(ctxGPG, PassphraseCallback, this);

  // Set GPGME's settings to match GPGee's configuration
  gpgme_control(ctxGPG, GPGME_CTRL_FORCETRUST, GPGeeConfig->Values["Force Trust"]?1:0);
  sConfigString = GPGeeConfig->Values["GPG Program"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_GPGPROGRAM, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Options File"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_OPTFILE, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_KEYRING, (int)sConfigString.c_str());
  sConfigString = GPGeeConfig->Values["Secret Keyring"];
  if (!sConfigString.IsEmpty())
    gpgme_control(ctxGPG, GPGME_CTRL_SECRETKEYRING, (int)sConfigString.c_str());
}  // __fastcall TformGPGeeVerifyDecrypt::TformGPGeeVerifyDecrypt(TComponent* Owner): TForm(Owner)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Death and destruction
//
__fastcall TformGPGeeVerifyDecrypt::~TformGPGeeVerifyDecrypt()
{
  // Clean up the lvFileList component - free all the data structures attached to each TListItem
  for (int n=0; n<lvFileList->Items->Count; n++) {
    delete (VerifyDecryptInfo_t *)lvFileList->Items->Item[n]->Data;
    lvFileList->Items->Item[n]->Data = NULL;
  }  // for (int n=0; n<lvFileList->Items->Count; n++)

  // Release any GPGME contexts that were created.
  if (cacheSecret)
    gpgme_keycache_release(cacheSecret);
  if (cachePublic)
    gpgme_keycache_release(cachePublic);
  if (ctxGPG)
    gpgme_release(ctxGPG);
}  // __fastcall TformGPGeeVerifyDecrypt::~TformGPGeeVerifyDecrypt()
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnShow event handler for the form.  The form has been fully streamed in
// by the time this is called, so we can do all the form's visual setup here.
//
void __fastcall TformGPGeeVerifyDecrypt::FormShow(TObject *Sender)
{
  TListItem *FileItem;

  // Sort the filenames, clear the lvFileList TListView and then add each file to the list, creating a
  // VerifyDecryptInfo_t structure for each one.  The structure holds the actual information and operations
  // results for each file.
  FileNames->Sort();
  lvFileList->Clear();
  for (int n=0; n<FileNames->Count; n++) {
    VerifyDecryptInfo_t *ItemInfo = new VerifyDecryptInfo_t;

    FileItem = lvFileList->Items->Add();
    FileItem->Caption = MinimizeName(FileNames->Strings[n],lvFileList->Font,lvFileList->Column[0]->Width-20);
    FileItem->Data = ItemInfo;

    ItemInfo->FileName = FileNames->Strings[n];
    ItemInfo->FileType = IdentifyFile(ItemInfo->FileName);
    if (ItemInfo->FileType == GPG_UNKNOWN) {
      ItemInfo->ResultString = GetMessage(MSG_ERROR_GPGEE_IDENTIFY);
      FileItem->StateIndex = GPGEE_STATE_INVALID;
      FileItem->SubItems->Add(GetMessage(MSG_GPGEE_TYPE_UNKNOWN));
    } else {
      FileItem->StateIndex = GPGEE_STATE_PENDING;
      FileItem->SubItems->Add((ItemInfo->FileType==GPG_MESSAGE)?GetMessage(MSG_GPGEE_TYPE_MESSAGE):GetMessage(MSG_GPGEE_TYPE_SIGNATURE));
    }  // if (ItemInfo->FileType == GPG_UNKNOWN)
  }  // for (int n=0; n<FileNames->Count; n++)
  lvFileList->SetFocus();
  StartTimer->Enabled = true;
}  // void __fastcall TformGPGeeVerifyDecrypt::FormShow(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Set the default modes and show the form.  Not much done here for this form
// except copy in the list of filenames we'll be working on.
//
int __fastcall TformGPGeeVerifyDecrypt::ShowModalWithDefaults(TStringList *FileNames)
{
  this->FileNames = FileNames;
  return ShowModal();
}  // int __fastcall TformGPGeeSignEncrypt::ShowModalWithDefaults(bool bSign, bool bEncrypt)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// StartTimer OnTimer event - this is used to allow the form time to get
// painted before we start into the operations.  This is set to 250ms in the
// form designer.
//
void __fastcall TformGPGeeVerifyDecrypt::StartTimerTimer(TObject *Sender)
{
  StartTimer->Enabled = false;
  VerifyDecryptFiles();
}  // void __fastcall TformGPGeeVerifyDecrypt::StartTimerTimer(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// The main work of verification/decryption is done here.
// The theory is simple - iterate through each entry in the lvFileList
// TListView compoment, figure out what type of file it is, perform the
// proper operation on it, and store the results for the user to check.
//
void __fastcall TformGPGeeVerifyDecrypt::VerifyDecryptFiles(void)
{
  gpgme_error_t err;
  gpgme_sig_t sig = NULL;

  // We want to make sure the form is drawn.  It should be by the time this method is called, but just in case
  // the system is really bogged, we do a ProcessMessages() right off the bat.  We also do one each time something
  // visual changes so that the form acts as its own progress bar, so to speak.  The user will be able to see as
  // each file is processed.
  Application->ProcessMessages();
  for (int n=0; n<lvFileList->Items->Count; n++) {
    TListItem *Item = lvFileList->Items->Item[n];
    VerifyDecryptInfo_t *ItemInfo = (VerifyDecryptInfo_t *)Item->Data;
    // This is the filename of the output file - simply the source filename without the .gpg/.sig/.asc extension
    AnsiString DataFileName = ItemInfo->FileName.SubString(0,ItemInfo->FileName.Length()-4);

    Item->Selected = true;                                      // Select each item as it's about to be processed
    if (Item->StateIndex == GPGEE_STATE_INVALID)                // Go on to the next file if this file is already marked as invalid
      continue;
    Item->StateIndex = GPGEE_STATE_INPROGRESS;                  // Otherwise mark the file as being in progress
    Application->ProcessMessages();                             // Make sure the user sees the in-progress icon
    if (Cancel) {                                               // If the cancel flag is set then all subsequent files
      Item->StateIndex = GPGEE_STATE_INVALID;                   // are marked as invalid.
      ItemInfo->ResultString = GetMessage(MSG_ERROR_GPGEE_CANCELED);
      continue;
    }  // if (Cancel)
    switch (ItemInfo->FileType) {
      case GPG_MESSAGE: {                                       // If this is "message" then we will decrypt it
        gpgme_op_flags_t OpFlags;                               // Result flags for the operation
        char KeyID[17];

        this->CurrentOpFileName = ItemInfo->FileName;               // Store the filename in case this file is going to be symmetrically decrypted so the the passphrase dialog can show the filename
        if (FileExists(DataFileName)) {                         // Figure out if we would be overwriting a file and then check if we're allowed to
          AnsiString FileOverwrite = GPGeeConfig->Values["File Overwrite"];
          if (FileOverwrite.AnsiCompareIC("never") == 0 || FileOverwrite.AnsiCompareIC("ask") == 0 &&
              Application->MessageBox(FormMessage(MSG_QUERY_GPGEE_OVERWRITE_MSG,DataFileName.c_str()).c_str(), GetMessage(MSG_QUERY_GPGEE_OVERWRITE_CAPTION).c_str(), MB_YESNO) == IDNO) {
            Item->StateIndex = GPGEE_STATE_INVALID;
            Application->ProcessMessages();
            ItemInfo->ResultString = GetMessage(MSG_ERROR_GPGEE_FILE_EXISTS);
            continue;
          }  // if (can't overwrite)
        }  // if (FileExists(DataFileName))
        // Now actually "decrypt" the file.
        err = gpgme_op_file_decrypt(ctxGPG, ("\"" + ItemInfo->FileName + "\"").c_str(), ("\"" + DataFileName + "\"").c_str());
        gpgme_decrypt_get_status(ctxGPG, KeyID, &OpFlags);
        if (err == GPGME_No_Error) {                            // If decrypting was successfil...
          ItemInfo->ResultString = FormMessage(MSG_RESULT_GPGEE_DECRYPT_SUCCESS,DataFileName.c_str());
          gpgme_decrypt_get_sig_ctx(ctxGPG, &sig);              // Check to see if the file we decrypted was signed
          // If the file was signed then we base the success of the operation on whether the signature matches.
          // Otherwise, if the file wasn't signed, then the operation was a success
          if (sig)
            Item->StateIndex = InterpretSignatureResult(sig, ItemInfo->ResultString);
          else
            Item->StateIndex = GPGEE_STATE_SUCCESS;
        } else {                                                // If decry[ting wasn't successful...
          ItemInfo->ResultString = String(gpgme_strerror(err));
          Item->StateIndex = GPGEE_STATE_FAILURE;
        }  // if (err == GPGME_No_Error)
        Application->ProcessMessages();                         // Make sure the user sees the state icon change
        // We check the operation's result flags to provide more information to the user.
        if (OpFlags & GPGME_OPFLAG_NOSECKEY)
          ItemInfo->ResultString += " " + FormMessage(MSG_RESULT_GPGEE_NO_SECRET_KEY,KeyID);
        if (OpFlags & GPGME_OPFLAG_BADARMOR)
          ItemInfo->ResultString += " " + GetMessage(MSG_RESULT_GPGEE_BAD_ARMOR);
        if (OpFlags & GPGME_OPFLAG_BADMDC)
          ItemInfo->ResultString += " " + GetMessage(MSG_RESULT_GPGEE_BAD_MDC);
        break;
      }  // case GPG_MESSAGE:
      case GPG_SIGNATURE: {                                     // If this is a detached signature then we verify it
        // If this is a detached signature then DataFileName isn't the output file but the name of the file that
        // was signed.  In this case the file must exist before the signature can be checked.
        if (!FileExists(DataFileName)) {
          Item->StateIndex = GPGEE_STATE_INVALID;
          Application->ProcessMessages();
          ItemInfo->ResultString = FormMessage(MSG_RESULT_GPGEE_NO_VERIFY_FILE,DataFileName.c_str());
          break;
        }  // if (!FileExists(DataFileName))
        // Verify the signature and then parse the results
        err = gpgme_op_file_verify(ctxGPG, GPGME_SIG_MODE_DETACH, &sig, ("\"" + ItemInfo->FileName + "\"").c_str(), ("\"" + DataFileName + "\"").c_str());
        Item->StateIndex = InterpretSignatureResult(sig, ItemInfo->ResultString, err);
        break;
      }  // case GPG_SIGNATURE:
    }  // switch (ItemInfo->FileType)
    if (sig) {
      // If our above operation ended up referencing a signature result context then release it and set the pointer
      // to NULL so we can tell for the next file if a signature result is referenced
      gpgme_sig_release(sig);
      sig = NULL;
    }  // if (sig)
  }  // for (int n=0; n<lvFileList->Items->Count; n++)
  lvFileList->Selected = lvFileList->Items->Item[0];              // Select the first item in the list when we're done.
  btnOkCancel->Caption = GetMessage(MSG_CAPTION_GPGEE_BUTTON_OK); // The Ok/Cancel button becomes the Ok button when all is done
  btnOkCancel->Tag = true;
}  // void __fastcall TformGPGeeVerifyDecrypt::VerifyDecryptFiles(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Look at the result of checking a signature and return the appropriate
// state code.  The state code returned is the same state code used to
// represent the icon state for the listview component.  In additaion,
// ResultString is appended with a text description of the signature result.
//
int __fastcall TformGPGeeVerifyDecrypt::InterpretSignatureResult(gpgme_sig_t sig, AnsiString &ResultString, gpgme_error_t err)
{
  AnsiString ResultStringPrefix, ResultStringBody, ResultStringPostfix;
  AnsiString ExtraResultString = "";
  AnsiString SigKeyId;                                          // ID of the signing key.

  int retval;

  if (!ResultString.IsEmpty())
    ResultString += "\r\n";                                     // Add a CRLF if the result string already has some stuff in it

  SigKeyId = String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_KEYID));
  // The basis for whether we consider the signature good or not is the signature's validity attribute.
  switch (gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_VALIDITY)) {
    case GPGME_SIG_STAT_NONE:                                   // No signature??
      ResultStringBody = FormMessage(MSG_RESULT_GPGEE_BODYNONE,gpgme_strerror(err));
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_GOOD:                                   // Good signature
      ResultStringPrefix = GetMessage(MSG_RESULT_GPGEE_PREFIXGOOD);
      retval = GPGEE_STATE_SUCCESS;
      break;
    case GPGME_SIG_STAT_BAD:                                    // Bad signature
      ResultStringPrefix = GetMessage(MSG_RESULT_GPGEE_PREFIXBAD);
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_NOKEY:                                  // Don't have key to check signature with
      ResultStringBody = FormMessage(MSG_RESULT_GPGEE_BODYNOKEY,SigKeyId.c_str());
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_NOSIG:                                  // No signature??
      ResultStringBody = GetMessage(MSG_RESULT_GPGEE_BODYNOSIG);
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_ERROR:                                  // Signature error of some sort?
      ResultStringBody = FormMessage(MSG_RESULT_GPGEE_BODYERROR,gpgme_strerror(err));
      retval = GPGEE_STATE_FAILURE;
      break;
    case GPGME_SIG_STAT_E_GOOD: {                               // Good signature with expired key - check dates on signature and key expiry
      AnsiString SigningKeyId;
      gpgme_key_t SigningKey;
      TDateTime KeyExpiryDate, SignatureCreationDate;

      ResultStringPrefix = GetMessage(MSG_RESULT_GPGEE_PREFIXEGOOD);
      CachePublicKeys();                                        // We need to cache the public keys so we can look up the signing key's expiry date
      // Here we check the expiry date of the signing key against the date the signature was created.  If the signature
      // was created before the key expired then we give a GPGEE_STATE_SUCCESS result (green checkmark).  If the signature
      // was created after the key expired (or if for some reason we run into a problem checking the dates) we give a
      // GPGEE_STATE_PARTIAL (yellow checkmark) result and warn the user.
      SigningKeyId = String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_KEYID)).SubString(9,8);
      err = gpgme_keycache_find_key(cachePublic, SigningKeyId.c_str(), false, &SigningKey);
      if (err == GPGME_No_Error) {
        // We don't convert dates from GMT to local time here.  The dates are left in GMT so there isn't any confusion
        // if the user needs to check with the document's signer about the expiry date on his/her key.  Any dates that
        // are presented to the user in GMT, though, are marked as such so that the user knows this.
        SignatureCreationDate = GPGToDateTime(gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_CREATED), false);
        KeyExpiryDate = GPGToDateTime(gpgme_key_get_ulong_attr(SigningKey, GPGME_ATTR_EXPIRES, NULL, 0), false);
        if (SignatureCreationDate > KeyExpiryDate) {
          ResultStringPostfix = FormMessage(MSG_RESULT_GPGEE_POSTEXPIREDBAD,DateTimeToStr(KeyExpiryDate).c_str());
          retval = GPGEE_STATE_PARTIAL;
        } else {
          ResultStringPostfix = GetMessage(MSG_RESULT_GPGEE_POSTEXPIREDOK);
          retval = GPGEE_STATE_SUCCESS;
        }  // if (SignatureCreationDate > KeyExpiryDate)
      } else {
        ResultStringPostfix = GetMessage(MSG_RESULT_GPGEE_POSTEXPIREDUNK);
        retval = GPGEE_STATE_PARTIAL;
      }  // if (err == GPGME_No_Error)
      break;
    }  // case GPGME_SIG_STAT_E_GOOD:
    case GPGME_SIG_STAT_R_GOOD:                                 // Good signature with revoked key
      ResultStringPrefix = GetMessage(MSG_RESULT_GPGEE_PREFIXGOOD);
      retval = GPGEE_STATE_PARTIAL;
      ResultStringPostfix = GetMessage(MSG_RESULT_GPGEE_POSTREVOKED);
      break;
  }  // switch (gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_VALIDITY))
  // We add information about the signature to some of the results.
  if (ResultStringBody.IsEmpty())
    ResultStringBody = " " + FormMessage(MSG_RESULT_GPGEE_BODYSTANDARD, gpgme_sig_get_string_attr(sig, GPGME_ATTR_USERID), KeyTypeToString((gpgme_pk_cipher_t)gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_ALGO)).c_str(), String(gpgme_sig_get_string_attr(sig, GPGME_ATTR_KEYID)).SubString(9,8).c_str(), DateTimeToStr(GPGToDateTime(gpgme_sig_get_ulong_attr(sig, 0, GPGME_ATTR_CREATED))).c_str());
  if (!ResultStringPostfix.IsEmpty())
    ResultStringPostfix = "\r\n" + ResultStringPostfix;
  ResultString += ResultStringPrefix + ResultStringBody + ResultStringPostfix;
  return retval;
}  // int __fastcall TformGPGeeVerifyDecrypt::InterpretSignatureResult(gpgme_sig_t sig, AnsiString &ResultString, gpgme_error_t err)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// OnChange event handler for the listview component.  Display the extended
// information about a file in the memo field when the user selects it.
//
void __fastcall TformGPGeeVerifyDecrypt::lvFileListChange(TObject *Sender, TListItem *Item, TItemChange Change)
{
  VerifyDecryptInfo_t *Info = (VerifyDecryptInfo_t *)Item->Data;
  if (Change == ctState && Info)
    richResults->Lines->Text = Info->ResultString;
}  // void __fastcall TformGPGeeVerifyDecrypt::lvFileListChange(TObject *Sender, TListItem *Item, TItemChange Change)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// User pressed the ok/cancel button - If we are currently performing
// operations then we set the cancel flag.  This will cause the rest of the
// operations to abort and leave the dialog up so the user can see the result
// of any completed operations.  If operations are completed when this is
// called, then the button has been changed to the "Ok" button and we simply
// close the dialog.
//
void __fastcall TformGPGeeVerifyDecrypt::btnOkCancelClick(TObject *Sender)
{
  if (btnOkCancel->Tag)
    ModalResult = mrOk;
  else
    Cancel = true;
}  // void __fastcall TformGPGeeVerifyDecrypt::btnOkCancelClick(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// User pressed the help button - pull up the correct help page
void __fastcall TformGPGeeVerifyDecrypt::btnHelpClick(TObject *Sender)
{
  Application->HelpContext(3);
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Make the buttons line up when the form changes size
//
void __fastcall TformGPGeeVerifyDecrypt::FormResize(TObject *Sender)
{
  btnOkCancel->Left = Width / 2 - btnOkCancel->Width / 2;
}  // void __fastcall TformGPGeeVerifyDecrypt::FormResize(TObject *Sender)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Cache the public keys.
// For verification and decryption, a public key cache is rarely needed, so
// it is only cached on demand.
//
void __fastcall TformGPGeeVerifyDecrypt::CachePublicKeys(void)
{
  gpgme_error_t err;

  if (!cachePublic) {
    err = gpgme_keycache_new(&cachePublic, ctxGPG);
    if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_INIT_PUB,gpgme_strerror(err)));
    err = gpgme_keycache_init(cachePublic, NULL, false);
    if (err)
    throw EGPGMEError(FormMessage(MSG_ERROR_GPGME_KEYCACHE_LOAD_PUB,gpgme_strerror(err)));
  }  // if (!cachePublic)
}  // void __fastcall TformGPGeeVerifyDecrypt::CachePublicKeys(void)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// This is where the real work of getting a passphrase is done.  It's easier
// to do this from inside the class, so the callback passes control here.
// When GPGME asks for a passphrase, it puts some information in a text
// string as to what sort of passphrase it is asking for.  It appears to
// always be three lines of text.  The second line contains a "User ID hint"
// which is the Key Id of the key for which the password is needed to unlock.
// If that Key Id is present it is assumed this is being called to unlock a
// secret key, and we get all the info on that key to present to the user.
// If that hint is absent then it is assumed this is being called to get a
// symmetrical-encryption passphrase.  OpenPGP smart cards are handled a
// little differently - I don't have one so I can't check on this yet.  It's
// likely that this won't work with them right now.
//
const char * __fastcall TformGPGeeVerifyDecrypt::GetPassphrase(const char *desc)
{
  static AnsiString Passphrase = "";
  gpgme_key_t Key;
  AnsiString KeyId, KeyName, KeyType, KeySize, KeyCreationDate;

  if (desc) {
    TStringList *desclines = new TStringList;
    formGPGeePassPhrase = new TformGPGeePassPhrase(NULL);

    // Drop the passphrase "description" text into a TStringList to parse it out into the individual lines
    desclines->Text = desc;
    // If there is no User ID hint then we assume that we are being invoked to get a symmetrical-encryption passphrase
    if (desclines->Strings[1] == "[User ID hint missing]") {
      if (formGPGeePassPhrase->GetPassphrase(FormMessage(MSG_QUERY_GPGEE_SYMPASSPHRASE,CurrentOpFileName.c_str()), "", "", "", "") == mrOk)
        Passphrase = formGPGeePassPhrase->edtPassphrase->Text;
      else
        Passphrase = "";
    // Otherwise, check line 2 of the description for the key id of the secret key we are getting the passphrase for and
    // present the user with the info about that key
    } else if (!desclines->Strings[1].IsEmpty()) {
      KeyId = desclines->Strings[1].SubString(9,8);
      gpgme_keycache_rewind(cacheSecret);
      if (gpgme_keycache_find_key(cacheSecret,KeyId.c_str(), false, &Key) == GPGME_No_Error) {
        KeyName = String(gpgme_key_get_string_attr(Key, GPGME_ATTR_NAME, NULL, 0)) + " <" + String(gpgme_key_get_string_attr(Key, GPGME_ATTR_EMAIL, NULL, 0)) + ">";
        KeyType = String(gpgme_key_get_string_attr(Key, GPGME_ATTR_ALGO, NULL, 0));
        KeySize = String(gpgme_key_get_ulong_attr(Key, GPGME_ATTR_LEN, NULL, 0));
        KeyCreationDate = DateToStr(GPGToDateTime(gpgme_key_get_ulong_attr(Key, GPGME_ATTR_CREATED, NULL, 0)));
        if (formGPGeePassPhrase->GetPassphrase(KeyName, KeyId, KeyType, KeySize, KeyCreationDate) == mrOk)
          Passphrase = formGPGeePassPhrase->edtPassphrase->Text;
        else
          Passphrase = "";
      }  // if (gpgme_keycache_find_key(cacheSecret,KeyId.c_str(), false, &Key) == GPGME_No_Error)
    }  // if (desclines->Strings[1] == "[User ID hint missing]")
    delete formGPGeePassPhrase;
    return Passphrase.c_str();
  // If we're called with an empty desc then we are being asked to clean up after a previous call (the passphrase is
  // no longer needed).  If our passphrase variable isn't empty then overwrite it before setting it to null length.
  } else if (Passphrase.IsEmpty()) {
    // Make sure that when the passphrase is no longer needed that it is erased
    Passphrase.StringOfChar('*', Passphrase.Length());
    Passphrase = "";
  }  // if (desc)
  return NULL;
}  // const char * __fastcall TformGPGeeVerifyDecrypt::GetPassphrase(const char *desc)
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// The actual passphrase callback - just dumps control to the Passphrase
// method of the form class.
//
const char * PassphraseCallback(void *hook, const char *desc, void *dummy)
{
  TformGPGeeVerifyDecrypt *formGPGeeVerifyDecrypt = (TformGPGeeVerifyDecrypt *)hook;
  return formGPGeeVerifyDecrypt->GetPassphrase(desc);
}
//---------------------------------------------------------------------------




